@@ -12,15 +12,11 @@
 //! real-number tokenisation.
 
 use hashbrown::HashMap;
+use pyo3::prelude::*;
 
 use std::path::Path;
 
-/// Types of failures from the lexer.
-#[derive(Clone, Copy, Debug)]
-pub enum LexerFailure {
-    FailedRead,
-    BadEncoding,
-}
+use crate::error::{message_generic, Position, QASM2ParseError};
 
 /// Tokenised version information data.  This is more structured than the real number suggested by
 /// the specification.
@@ -122,11 +118,51 @@ pub enum TokenType {
     Integer,
     Filename,
     Version,
-    // Not actually a valid token, used to more easily signal an error.
-    Error,
 }
 
 impl TokenType {
+    pub fn variable_text(&self) -> bool {
+        match self {
+            TokenType::OpenQASM
+            | TokenType::Barrier
+            | TokenType::Cos
+            | TokenType::Creg
+            | TokenType::Exp
+            | TokenType::Gate
+            | TokenType::If
+            | TokenType::Include
+            | TokenType::Ln
+            | TokenType::Measure
+            | TokenType::Opaque
+            | TokenType::Qreg
+            | TokenType::Reset
+            | TokenType::Sin
+            | TokenType::Sqrt
+            | TokenType::Tan
+            | TokenType::Pi
+            | TokenType::Plus
+            | TokenType::Minus
+            | TokenType::Arrow
+            | TokenType::Asterisk
+            | TokenType::Equals
+            | TokenType::Slash
+            | TokenType::Caret
+            | TokenType::Semicolon
+            | TokenType::Comma
+            | TokenType::LParen
+            | TokenType::RParen
+            | TokenType::LBracket
+            | TokenType::RBracket
+            | TokenType::LBrace
+            | TokenType::RBrace => false,
+            TokenType::Id
+            | TokenType::Real
+            | TokenType::Integer
+            | TokenType::Filename
+            | TokenType::Version => true,
+        }
+    }
+
     /// Get a static description of the token type.  This is useful for producing messages when the
     /// full token context isn't available, or isn't important.
     pub fn describe(&self) -> &'static str {
@@ -168,7 +204,6 @@ impl TokenType {
             TokenType::Integer => "an integer",
             TokenType::Filename => "a filename string",
             TokenType::Version => "a '<major>.<minor>' version",
-            TokenType::Error => "<invalid token during lexing>",
         }
     }
 }
@@ -191,19 +226,6 @@ pub struct Token {
 }
 
 impl Token {
-    /// Create a dummy token that doesn't represent anything.  This is useful in a couple of places
-    /// where the general functions of the parser always take a token to "blame" if something goes
-    /// wrong, but the operation is being called for built-in objects (like the `U` gate) that are
-    /// guaranteed to succeed.
-    pub fn dummy() -> Self {
-        Token {
-            ttype: TokenType::Error,
-            line: 0,
-            col: 0,
-            index: usize::MAX,
-        }
-    }
-
     /// Get a reference to the string that was seen to generate this token.
     pub fn text<'a>(&self, context: &'a TokenContext) -> &'a str {
         match self.ttype {
@@ -289,6 +311,7 @@ pub struct TokenStream {
     /// The filename from which this stream is derived.  May be a placeholder if there is no
     /// backing file or other named resource.
     pub filename: std::ffi::OsString,
+    strict: bool,
     source: Box<dyn std::io::BufRead + Send>,
     line_buffer: Vec<u8>,
     done: bool,
@@ -303,11 +326,17 @@ pub struct TokenStream {
 impl TokenStream {
     /// Create and initialise a generic [TokenStream], given a source that implements
     /// [std::io::BufRead] and a filename (or resource path) that describes its source.
-    fn new(source: Box<dyn std::io::BufRead + Send>, filename: std::ffi::OsString) -> Self {
+    fn new(
+        source: Box<dyn std::io::BufRead + Send>,
+        filename: std::ffi::OsString,
+        strict: bool,
+    ) -> Self {
         TokenStream {
+            filename,
+            strict,
             source,
             line_buffer: Vec::with_capacity(80),
-            filename,
+            done: false,
             // The first line is numbered "1", and the first column is "0".  The counts are
             // initialised like this so the first call to `next_byte` can easily detect that it
             // needs to extract the next line.
@@ -315,27 +344,31 @@ impl TokenStream {
             col: 0,
             try_version: false,
             peeked: None,
-            done: false,
         }
     }
 
     /// Create a [TokenStream] from a string containing the OpenQASM 2 program.
-    pub fn from_string(string: String) -> Self {
-        TokenStream::new(Box::new(std::io::Cursor::new(string)), "<input>".into())
+    pub fn from_string(string: String, strict: bool) -> Self {
+        TokenStream::new(
+            Box::new(std::io::Cursor::new(string)),
+            "<input>".into(),
+            strict,
+        )
     }
 
     /// Create a [TokenStream] from a path containing the OpenQASM 2 program.
-    pub fn from_path<P: AsRef<Path>>(path: P) -> Result<Self, std::io::Error> {
+    pub fn from_path<P: AsRef<Path>>(path: P, strict: bool) -> Result<Self, std::io::Error> {
         let file = std::fs::File::open(path.as_ref())?;
         Ok(TokenStream::new(
             Box::new(std::io::BufReader::new(file)),
             Path::file_name(path.as_ref()).unwrap().into(),
+            strict,
         ))
     }
 
     /// Read the next line into the managed buffer in the struct, updating the tracking information
     /// of the position, and the `done` state of the iterator.
-    fn advance_line(&mut self) -> Result<usize, LexerFailure> {
+    fn advance_line(&mut self) -> PyResult<usize> {
         if self.done {
             Ok(0)
         } else {
@@ -351,9 +384,12 @@ impl TokenStream {
                     }
                     Ok(count)
                 }
-                Err(_) => {
+                Err(err) => {
                     self.done = true;
-                    Err(LexerFailure::FailedRead)
+                    Err(QASM2ParseError::new_err(message_generic(
+                        Some(&Position::new(&self.filename, self.line, self.col)),
+                        &format!("lexer failed to read stream: {}", err),
+                    )))
                 }
             }
         }
@@ -361,16 +397,19 @@ impl TokenStream {
 
     /// Get the next character in the stream.  This updates the line and column information for the
     /// current bute as well.
-    fn next_byte(&mut self) -> Result<Option<u8>, LexerFailure> {
+    fn next_byte(&mut self) -> PyResult<Option<u8>> {
         if self.col >= self.line_buffer.len() && self.advance_line()? == 0 {
             return Ok(None);
         }
         let out = self.line_buffer[self.col];
         self.col += 1;
         match out {
-            0x80..=0xff => {
+            b @ 0x80..=0xff => {
                 self.done = true;
-                Err(LexerFailure::BadEncoding)
+                Err(QASM2ParseError::new_err(message_generic(
+                    Some(&Position::new(&self.filename, self.line, self.col)),
+                    &format!("encountered a non-ASCII byte: {:02X?}", b),
+                )))
             }
             b => Ok(Some(b)),
         }
@@ -379,27 +418,48 @@ impl TokenStream {
     /// Peek at the next byte in the stream without consuming it.  This still returns an error if
     /// the next byte isn't in the valid range for OpenQASM 2, or if the file/stream has failed to
     /// read into the buffer for some reason.
-    fn peek_byte(&mut self) -> Result<Option<u8>, LexerFailure> {
+    fn peek_byte(&mut self) -> PyResult<Option<u8>> {
         if self.col >= self.line_buffer.len() && self.advance_line()? == 0 {
             return Ok(None);
         }
         match self.line_buffer[self.col] {
-            0x80..=0xff => {
+            b @ 0x80..=0xff => {
                 self.done = true;
-                Err(LexerFailure::BadEncoding)
+                Err(QASM2ParseError::new_err(message_generic(
+                    Some(&Position::new(&self.filename, self.line, self.col)),
+                    &format!("encountered a non-ASCII byte: {:02X?}", b),
+                )))
             }
             b => Ok(Some(b)),
         }
     }
 
+    /// Expect that the next byte is not a word continuation, providing a suitable error message if
+    /// it is.
+    fn expect_word_boundary(&mut self, after: &str, start_col: usize) -> PyResult<()> {
+        match self.peek_byte()? {
+            Some(c @ (b'a'..=b'z' | b'A'..=b'Z' | b'0'..=b'9' | b'_')) => {
+                Err(QASM2ParseError::new_err(message_generic(
+                    Some(&Position::new(&self.filename, self.line, start_col)),
+                    &format!(
+                        "expected a word boundary after {}, but saw '{}'",
+                        after, c as char
+                    ),
+                )))
+            }
+            _ => Ok(()),
+        }
+    }
+
     /// Complete the lexing of a floating point value after the decimal point has been consumed.
-    fn lex_float_after_decimal(&mut self) -> Result<(TokenType, bool), LexerFailure> {
+    fn lex_float_after_decimal(&mut self, start_col: usize) -> PyResult<TokenType> {
         // Consume the rest of the fractional part.
         while let Some(b'0'..=b'9') = self.peek_byte()? {
             self.next_byte()?;
         }
         if !matches!(self.peek_byte()?, Some(b'e' | b'E')) {
-            return Ok((TokenType::Real, true));
+            self.expect_word_boundary("a float", start_col)?;
+            return Ok(TokenType::Real);
         }
         // Consume the rest of the exponent.
         self.next_byte()?;
@@ -408,25 +468,31 @@ impl TokenStream {
         }
         // Exponents must have at least one digit in them.
         if !matches!(self.peek_byte()?, Some(b'0'..=b'9')) {
-            return Ok((TokenType::Error, false));
+            return Err(QASM2ParseError::new_err(message_generic(
+                Some(&Position::new(&self.filename, self.line, start_col)),
+                "needed to see an integer exponent for this float",
+            )));
         }
         while let Some(b'0'..=b'9') = self.peek_byte()? {
             self.next_byte()?;
         }
-        Ok((TokenType::Real, true))
+        self.expect_word_boundary("a float", start_col)?;
+        Ok(TokenType::Real)
     }
 
     /// Lex a numeric token completely.  This can return a successful integer or a real number; the
     /// function distinguishes based on what it sees.
-    fn lex_numeric(&mut self) -> Result<(TokenType, bool), LexerFailure> {
-        let first_index = self.col - 1;
-        let first = self.line_buffer[first_index];
+    fn lex_numeric(&mut self, start_col: usize) -> PyResult<TokenType> {
+        let first = self.line_buffer[start_col];
         if first == b'.' {
             return match self.next_byte()? {
                 // In the case of a float that begins with '.', we require at least one digit, so
                 // just force consume it and continue like normal.
-                Some(b'0'..=b'9') => self.lex_float_after_decimal(),
-                _ => Ok((TokenType::Error, false)),
+                Some(b'0'..=b'9') => self.lex_float_after_decimal(start_col),
+                _ => Err(QASM2ParseError::new_err(message_generic(
+                    Some(&Position::new(&self.filename, self.line, start_col)),
+                    "expected a numeric fractional part after the bare decimal point",
+                ))),
             };
         }
         while let Some(b'0'..=b'9') = self.peek_byte()? {
@@ -434,82 +500,98 @@ impl TokenStream {
         }
         if let Some(b'.') = self.peek_byte()? {
             self.next_byte()?;
-            return self.lex_float_after_decimal();
+            return self.lex_float_after_decimal(start_col);
         }
-        if first == b'0' && self.col - first_index > 1 {
+        if first == b'0' && self.col - start_col > 1 {
             // Integers can't start with a leading zero unless they are only the single '0', but we
             // didn't see a decimal point.
-            Ok((TokenType::Error, false))
+            Err(QASM2ParseError::new_err(message_generic(
+                Some(&Position::new(&self.filename, self.line, start_col)),
+                "integers cannot have leading zeroes",
+            )))
         } else {
-            Ok((TokenType::Integer, true))
+            self.expect_word_boundary("an integer", start_col)?;
+            Ok(TokenType::Integer)
         }
     }
 
     /// Lex a text-like token into a complete token.  This can return any of the keyword-like
     /// tokens (e.g. [TokenType::Pi]), or a [TokenType::Id] if the token is not a built-in keyword.
-    fn lex_textlike(&mut self) -> Result<(TokenType, bool), LexerFailure> {
-        let first_index = self.col - 1;
-        let first = self.line_buffer[first_index];
+    fn lex_textlike(&mut self, start_col: usize) -> PyResult<TokenType> {
+        let first = self.line_buffer[start_col];
         while let Some(b'a'..=b'z' | b'A'..=b'Z' | b'0'..=b'9' | b'_') = self.peek_byte()? {
             self.next_byte()?;
         }
-        let text = &self.line_buffer[first_index..self.col];
+        // No need to expect the word boundary after this, because it's the same check as above.
+        let text = &self.line_buffer[start_col..self.col];
         if let b'A'..=b'Z' = first {
             match text {
                 b"OPENQASM" => {
                     self.try_version = true;
-                    Ok((TokenType::OpenQASM, false))
+                    Ok(TokenType::OpenQASM)
                 }
-                b"U" | b"CX" => Ok((TokenType::Id, true)),
-                _ => Ok((TokenType::Error, false)),
+                b"U" | b"CX" => Ok(TokenType::Id),
+                _ => Err(QASM2ParseError::new_err(message_generic(
+                        Some(&Position::new(&self.filename, self.line, start_col)),
+                        "identifiers cannot start with capital letters except for the builtins 'U' and 'CX'"))),
             }
         } else {
             match text {
-                b"barrier" => Ok((TokenType::Barrier, false)),
-                b"cos" => Ok((TokenType::Cos, false)),
-                b"creg" => Ok((TokenType::Creg, false)),
-                b"exp" => Ok((TokenType::Exp, false)),
-                b"gate" => Ok((TokenType::Gate, false)),
-                b"if" => Ok((TokenType::If, false)),
-                b"include" => Ok((TokenType::Include, false)),
-                b"ln" => Ok((TokenType::Ln, false)),
-                b"measure" => Ok((TokenType::Measure, false)),
-                b"opaque" => Ok((TokenType::Opaque, false)),
-                b"qreg" => Ok((TokenType::Qreg, false)),
-                b"reset" => Ok((TokenType::Reset, false)),
-                b"sin" => Ok((TokenType::Sin, false)),
-                b"sqrt" => Ok((TokenType::Sqrt, false)),
-                b"tan" => Ok((TokenType::Tan, false)),
-                b"pi" => Ok((TokenType::Pi, false)),
-                _ => Ok((TokenType::Id, true)),
+                b"barrier" => Ok(TokenType::Barrier),
+                b"cos" => Ok(TokenType::Cos),
+                b"creg" => Ok(TokenType::Creg),
+                b"exp" => Ok(TokenType::Exp),
+                b"gate" => Ok(TokenType::Gate),
+                b"if" => Ok(TokenType::If),
+                b"include" => Ok(TokenType::Include),
+                b"ln" => Ok(TokenType::Ln),
+                b"measure" => Ok(TokenType::Measure),
+                b"opaque" => Ok(TokenType::Opaque),
+                b"qreg" => Ok(TokenType::Qreg),
+                b"reset" => Ok(TokenType::Reset),
+                b"sin" => Ok(TokenType::Sin),
+                b"sqrt" => Ok(TokenType::Sqrt),
+                b"tan" => Ok(TokenType::Tan),
+                b"pi" => Ok(TokenType::Pi),
+                _ => Ok(TokenType::Id),
             }
         }
     }
 
     /// Lex a filename token completely.  This is always triggered by seeing a `b'"'` byte in the
     /// input stream.
-    fn lex_filename(&mut self) -> Result<(TokenType, bool), LexerFailure> {
+    fn lex_filename(&mut self, terminator: u8, start_col: usize) -> PyResult<TokenType> {
         loop {
             match self.next_byte()? {
-                None | Some(b'\n') | Some(b'\r') => return Ok((TokenType::Error, false)),
-                Some(b'"') => {
-                    return Ok((TokenType::Filename, true));
+                None => {
+                    return Err(QASM2ParseError::new_err(message_generic(
+                        Some(&Position::new(&self.filename, self.line, start_col)),
+                        "unexpected end-of-file while lexing string literal",
+                    )))
+                }
+                Some(b'\n' | b'\r') => {
+                    return Err(QASM2ParseError::new_err(message_generic(
+                        Some(&Position::new(&self.filename, self.line, start_col)),
+                        "unexpected line break while lexing string literal",
+                    )))
+                }
+                Some(c) if c == terminator => {
+                    return Ok(TokenType::Filename);
                 }
                 Some(_) => (),
             }
         }
     }
 
-    /// Attempt to lex a version identifier as the next some.  Returns the `Some` variant with the
-    /// token information if the version information matches correctly, and the `None` variant if
-    /// it doesn't.  This function should only be called immediately after the emission of a
+    /// Attempt to lex a version identifier as the next some.  Returns `true` if a version matched,
+    /// and `false` if not. This function should only be called immediately after the emission of a
     /// [TokenType::OpenQASM] token; it is not valid anywhere else in an OpenQASM 2 program.
     #[cold]
-    fn try_lex_version(&mut self) -> Result<Option<(TokenType, bool)>, LexerFailure> {
+    fn try_lex_version(&mut self) -> PyResult<bool> {
         if let Some(b'1'..=b'9') = self.peek_byte()? {
             self.next_byte()?;
         } else {
-            return Ok(None);
+            return Ok(false);
         }
         while let Some(b'0'..=b'9') = self.peek_byte()? {
             self.next_byte()?;
@@ -517,44 +599,32 @@ impl TokenStream {
         if let Some(b'.') = self.peek_byte()? {
             self.next_byte()?;
         } else {
-            return Ok(Some((TokenType::Version, true)));
+            return Ok(true);
         };
         // If we don't see at least one digit following the dot, it's still not a valid version.
         if let Some(b'0'..=b'9') = self.peek_byte()? {
         } else {
-            return Ok(None);
+            return Ok(false);
         }
         while let Some(b'0'..=b'9') = self.peek_byte()? {
             self.next_byte()?;
         }
-        Ok(Some((TokenType::Version, true)))
-    }
-
-    /// Create a new token representing an error in the stream.  This is a somewhat standardised
-    /// form, and potentially returned from a few places.
-    fn error_token(&self) -> Token {
-        Token {
-            ttype: TokenType::Error,
-            line: self.line,
-            col: self.col,
-            index: usize::MAX,
-        }
+        Ok(true)
     }
 
     /// The actual core of the iterator.  Read from the stream (ignoring preceding whitespace)
     /// until a complete [Token] has been constructed, or the end of the iterator is reached.  This
     /// returns `Some` for all tokens, including the error token, and only returns `None` if there
     /// are no more tokens left to take.
-    fn next_inner(&mut self, context: &mut TokenContext) -> Option<Token> {
+    fn next_inner(&mut self, context: &mut TokenContext) -> PyResult<Option<Token>> {
         // Consume preceding whitespace.  Beware that this can still exhaust the underlying stream,
         // or scan through an invalid token in the encoding.
         loop {
-            match self.peek_byte() {
-                Ok(None) => return None,
-                Ok(Some(b' ' | b'\t' | b'\r' | b'\n')) => {
-                    self.next_byte().ok()?;
+            match self.peek_byte()? {
+                Some(b' ' | b'\t' | b'\r' | b'\n') => {
+                    self.next_byte()?;
                 }
-                Err(_) => return Some(self.error_token()),
+                None => return Ok(None),
                 _ => break,
             }
         }
@@ -562,99 +632,104 @@ impl TokenStream {
         // This should quickly get branch-predicted to be very cold.  Even if the version fails to
         // match, we shouldn't try again, because another [Token] will get emitted instead, and
         // then we're no longer in the `OPENQASM` statement mode.
-        if self.try_version {
-            self.try_version = false;
-            match self.try_lex_version() {
-                Ok(Some((ttype, has_text))) => {
-                    return Some(Token {
-                        ttype,
-                        line: self.line,
-                        col: start_col,
-                        index: if has_text {
-                            context.index(&self.line_buffer[start_col..self.col])
-                        } else {
-                            usize::MAX
-                        },
-                    });
-                }
-                Err(_) => return Some(self.error_token()),
-                _ => (),
-            }
+        if self.try_version && self.try_lex_version()? {
+            return Ok(Some(Token {
+                ttype: TokenType::Version,
+                line: self.line,
+                col: start_col,
+                index: context.index(&self.line_buffer[start_col..self.col]),
+            }));
         }
+        self.try_version = false;
         // The whitespace loop (or [Self::try_lex_version]) has already peeked the next token, so
         // we know it's going to be the `Some` variant.
-        let (ttype, has_text) = match self.next_byte().map(Option::unwrap) {
-            Ok(b'+') => (TokenType::Plus, false),
-            Ok(b'*') => (TokenType::Asterisk, false),
-            Ok(b'^') => (TokenType::Caret, false),
-            Ok(b';') => (TokenType::Semicolon, false),
-            Ok(b',') => (TokenType::Comma, false),
-            Ok(b'(') => (TokenType::LParen, false),
-            Ok(b')') => (TokenType::RParen, false),
-            Ok(b'[') => (TokenType::LBracket, false),
-            Ok(b']') => (TokenType::RBracket, false),
-            Ok(b'{') => (TokenType::LBrace, false),
-            Ok(b'}') => (TokenType::RBrace, false),
-            Ok(b'/') => {
-                if let Ok(Some(b'/')) = self.peek_byte() {
-                    if self.advance_line().is_ok() {
-                        return self.next(context);
-                    } else {
-                        return Some(self.error_token());
-                    }
+        let ttype = match self.next_byte()?.unwrap() {
+            b'+' => TokenType::Plus,
+            b'*' => TokenType::Asterisk,
+            b'^' => TokenType::Caret,
+            b';' => TokenType::Semicolon,
+            b',' => TokenType::Comma,
+            b'(' => TokenType::LParen,
+            b')' => TokenType::RParen,
+            b'[' => TokenType::LBracket,
+            b']' => TokenType::RBracket,
+            b'{' => TokenType::LBrace,
+            b'}' => TokenType::RBrace,
+            b'/' => {
+                if let Some(b'/') = self.peek_byte()? {
+                    self.advance_line()?;
+                    return self.next(context);
                 } else {
-                    (TokenType::Slash, false)
+                    TokenType::Slash
                 }
             }
-            Ok(b'-') => {
+            b'-' => {
                 if let Ok(Some(b'>')) = self.peek_byte() {
                     self.col += 1;
-                    (TokenType::Arrow, false)
+                    TokenType::Arrow
                 } else {
-                    (TokenType::Minus, false)
+                    TokenType::Minus
                 }
             }
-            Ok(b'=') => {
+            b'=' => {
                 if let Ok(Some(b'=')) = self.peek_byte() {
                     self.col += 1;
-                    (TokenType::Equals, false)
+                    TokenType::Equals
                 } else {
-                    return Some(self.error_token());
+                    return Err(QASM2ParseError::new_err(
+                        "single equals '=' is never valid".to_owned(),
+                    ));
                 }
             }
-            Ok(b'0'..=b'9' | b'.') => self.lex_numeric().unwrap_or((TokenType::Error, false)),
-            Ok(b'a'..=b'z' | b'A'..=b'Z') => {
-                self.lex_textlike().unwrap_or((TokenType::Error, false))
+            b'0'..=b'9' | b'.' => self.lex_numeric(start_col)?,
+            b'a'..=b'z' | b'A'..=b'Z' => self.lex_textlike(start_col)?,
+            c @ (b'"' | b'\'') => {
+                if self.strict && c != b'"' {
+                    return Err(QASM2ParseError::new_err(message_generic(
+                        Some(&Position::new(&self.filename, self.line, start_col)),
+                        "[strict] paths must be in double quotes (\"\")",
+                    )));
+                } else {
+                    self.lex_filename(c, start_col)?
+                }
+            }
+            c => {
+                return Err(QASM2ParseError::new_err(message_generic(
+                    Some(&Position::new(&self.filename, self.line, start_col)),
+                    &format!(
+                        "encountered '{}', which doesn't match any valid tokens",
+                        // Non-ASCII bytes should already have been rejected by `next_byte()`.
+                        c as char,
+                    ),
+                )));
             }
-            Ok(b'"') => self.lex_filename().unwrap_or((TokenType::Error, false)),
-            _ => return Some(self.error_token()),
         };
-        Some(Token {
+        Ok(Some(Token {
             ttype,
             line: self.line,
             col: start_col,
-            index: if has_text {
+            index: if ttype.variable_text() {
                 context.index(&self.line_buffer[start_col..self.col])
             } else {
                 usize::MAX
             },
-        })
+        }))
     }
 
     /// Get an optional reference to the next token in the iterator stream without consuming it.
     /// This is a direct analogue of the same method on the [std::iter::Peekable] struct, except it
     /// is manually defined here to avoid hiding the rest of the public fields of the [TokenStream]
     /// struct itself.
-    pub fn peek(&mut self, context: &mut TokenContext) -> Option<&Token> {
+    pub fn peek(&mut self, context: &mut TokenContext) -> PyResult<Option<&Token>> {
         if self.peeked.is_none() {
-            self.peeked = Some(self.next_inner(context));
+            self.peeked = Some(self.next_inner(context)?);
         }
-        self.peeked.as_ref().unwrap().as_ref()
+        Ok(self.peeked.as_ref().unwrap().as_ref())
     }
 
-    pub fn next(&mut self, context: &mut TokenContext) -> Option<Token> {
+    pub fn next(&mut self, context: &mut TokenContext) -> PyResult<Option<Token>> {
         match self.peeked.take() {
-            Some(token) => token,
+            Some(token) => Ok(token),
             None => self.next_inner(context),
         }
     }
